/**
 * CS3725 - Assignment #2
 * Tim Oram (200529220)
 * 
 * License:
 * The MIT License (MIT) Copyright (c) 2013 Tim Oram
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this
 * software and associated documentation files (the "Software"), to deal in the Software
 * without restriction, including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons
 * to whom the Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all copies or
 * substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */
package ca.mitmaro.cs3725.a2;

import java.util.ArrayList;

import ca.mitmaro.cs3725.a2.machine.Cyclable;
import ca.mitmaro.cs3725.a2.machine.component.Component;
import ca.mitmaro.cs3725.a2.machine.component.DUnit;
import ca.mitmaro.cs3725.a2.machine.component.InstructionUnit;
import ca.mitmaro.cs3725.a2.machine.component.OperableComponent;
import ca.mitmaro.cs3725.a2.machine.component.RegisterBank;
import ca.mitmaro.cs3725.a2.program.Instruction;
import ca.mitmaro.cs3725.a2.program.Operation;

public class Machine {
	
	private RegisterBank registerBank;
	
	private InstructionUnit iu;
	
	private ArrayList<DUnit> dunits;
	
	private ArrayList<OperableComponent> components;
	
	private Program program;
	
	// no operation dunit, signifies a dunit that skipped a cycle
	private DUnit noop;
	
	public Machine(InstructionUnit iu, RegisterBank registerBank) {
		this.registerBank = registerBank;
		this.iu = iu;
		this.dunits = new ArrayList<DUnit>();
		this.noop = new DUnit(new Operation(""), this.registerBank, 0);
		this.components = new ArrayList<OperableComponent>();
	}
	
	// adds a dunit of operation type
	public void addDUnit(Operation operation, int cycles) {
		DUnit dunit = new DUnit(operation, this.registerBank, cycles);
		this.dunits.add(dunit);
		this.components.add(dunit);
	}
	
	// remove all dunits
	public void clearDUnits() {
		this.dunits.clear();
	}
	
	// return a dunit thats not busy and has the right operation
	public DUnit getFreeDUnit(Operation op) {
		for (DUnit d: this.dunits) {
			if ((d.isFree() || d.isInactive()) && d.equals(op)) {
				return d;
			}
		}
		return null;
	}
	
	// load a program into memory
	public void loadProgram(Program program) {
		this.program = program;
	}

	// this does all the hard work. Does a cycle of the machine
	public ArrayList<OperableComponent> cycle() {
		ArrayList<OperableComponent> operations = new ArrayList<OperableComponent>();
		
		Instruction inst;
		DUnit dunit;
		
		// free any blocked dunits
		for (DUnit du: this.dunits) {
			if (du.isBlocked()) {
				du.free();
			}
		}
		
		// release conflicted dunits that are now free
		for (DUnit du: this.dunits) {
			du.tryReleaseFromConflict();
		}
		
		// load instruction if IU is done
		if (this.iu.isBlocked()) {
			inst = this.iu.getInstruction();
			dunit = this.getFreeDUnit(inst.getOperation());
			if (dunit != null) { // dunit found
				dunit.reset();
				dunit.start(inst);
				this.iu.reset();
			}
		}
		
		// if IU is still free then we start loading the next instruction
		if (this.iu.isFree() && this.program.hasNext()) {
			this.iu.start(this.program.next());
			operations.add(iu);
		} else if (this.iu.isActive()) {
			operations.add(iu);
		}
		

		// loop over dunits adding them as a operation
		for (DUnit du: this.dunits) {
			if (du.isActive()) {
				operations.add(du);
			} else if (du.isConflict()) { // conflicting dunits have a noop
				this.noop.setInstruction(du.getInstruction()); // set instruction to fix order bug
				operations.add(this.noop);
			}
		}
		
		// cycle all components
		for (Cyclable c: this.components) {
			c.cycle();
		}
		this.iu.cycle();
		
		return operations;
	}
	
	// resets the machine back to defaults
	public void reset() {
		for (Component c: this.components) {
			c.reset();
		}
		this.registerBank.reset();
		this.program.reset();
	}
	
	// returns the program
	public Program getProgram() {
		return this.program;
	}
	
}
