/**
 * CS3725 - Assignment #3
 * Tim Oram (200529220)
 * 
 * License:
 * The MIT License (MIT) Copyright (c) 2013 Tim Oram
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this
 * software and associated documentation files (the "Software"), to deal in the Software
 * without restriction, including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons
 * to whom the Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all copies or
 * substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */
package ca.mitmaro.cs3725.a3;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;

import ca.mitmaro.cs3725.a3.machine.CycleOperations;
import ca.mitmaro.cs3725.a3.machine.component.InstructionUnit;
import ca.mitmaro.cs3725.a3.machine.component.OperableComponent;
import ca.mitmaro.cs3725.a3.machine.component.RegisterBank;
import ca.mitmaro.cs3725.a3.program.Operation;
import ca.mitmaro.prattle.Broker;

public class Application {
	
	// a runnable wrapper
	private Runnable runnable = new Runnable() {
		@Override
		public void run() {
			Application.this.next();
		}
	};

	private Machine machine;
	
	private Broker broker;
	
	private ScheduledExecutorService executorService;
	
	private ScheduledFuture<?> executor;
	
	public Application(Broker broker) {
		this.broker = broker;
		this.broker.registerTopic(new String[] {
			Topics.Application.reset,
			Topics.Application.cycleComplete,
			Topics.Application.programLoad,
			Topics.Application.programStop,
			Topics.Application.programStart,
			Topics.Application.programRun
		});
		
		this.executorService = Executors.newScheduledThreadPool(1);
	}
	
	// creates a machine with iu cycle time and number of registers
	public void createMachine(int iucycles, int registers) {
		this.machine = new Machine(new InstructionUnit(iucycles), new RegisterBank(registers));
	}
	
	// add a number of dunits to the machine
	public void addDUnits(String operation, int cycles, int number) {
		for (int i = 0; i < number; i++) {
			this.machine.addDUnit(new Operation(operation), cycles);
		}
	}
	
	// reset the application back to starting state
	public void reset() {
		this.stop();
		this.machine.reset();
		this.broker.post(Topics.Application.reset);
	}
	
	// start the application
	public void start() {
		
		HashMap<String, Object> context = new HashMap<String, Object>();
		context.put("registers", this.machine.getRegisters());

		this.broker.post(Topics.Application.programStart, context);
	}

	// start autorun
	public void autoRun(int timeDelay) {
		this.broker.post(Topics.Application.programRun);
		this.executor = this.executorService.scheduleAtFixedRate(this.runnable, 0, timeDelay, TimeUnit.MILLISECONDS);
	}

	// stop the auto run
	public void stop() {
		if (this.executor != null && !this.executor.isDone()) {
			this.executor.cancel(false);
			this.executor = null;
			this.broker.post(Topics.Application.programStop);
		}
	}
	
	// perform next cycle
	public void next() {
		
		// cycle the machine
		ArrayList<OperableComponent> ops = this.machine.cycle();

		// no operations then program is done 
		if (ops.size() == 0) {
			this.stop();
			this.reset();
			return;
		}
		
		// build cycle ops for gui (more or less)
		CycleOperations cycleOps = new CycleOperations();
		for (OperableComponent com: ops) {
			cycleOps.addOperation(com);
		}
		
		HashMap<String, Object> context = new HashMap<String, Object>();
		context.put("cycle", cycleOps);
		context.put("registers", this.machine.getRegisters());

		this.broker.post(Topics.Application.cycleComplete, context);

	}
	
	// load a program into machine memory from a text string
	public void loadProgram(String programText) throws IOException {
		
		Program program = new Program();
		
		for (String line: programText.split("\\n")) {
			program.add(line.trim());
		}
		
		this.machine.loadProgram(program);
		this.broker.post(Topics.Application.programLoad);
	}
	
	public Machine getMachine() {
		return this.machine;
	}

}
