/**
 * CS3725 - Assignment #3
 * Tim Oram (200529220)
 * 
 * License:
 * The MIT License (MIT) Copyright (c) 2013 Tim Oram
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this
 * software and associated documentation files (the "Software"), to deal in the Software
 * without restriction, including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons
 * to whom the Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all copies or
 * substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */
package ca.mitmaro.cs3725.a3;

import java.util.ArrayList;
import java.util.Collection;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ca.mitmaro.cs3725.a3.machine.ScoreBoard;
import ca.mitmaro.cs3725.a3.machine.component.DUnit;
import ca.mitmaro.cs3725.a3.machine.component.InstructionUnit;
import ca.mitmaro.cs3725.a3.machine.component.OperableComponent;
import ca.mitmaro.cs3725.a3.machine.component.Register;
import ca.mitmaro.cs3725.a3.machine.component.RegisterBank;
import ca.mitmaro.cs3725.a3.program.Instruction;
import ca.mitmaro.cs3725.a3.program.Operation;

public class Machine {
	
	private static final Logger logger = LoggerFactory.getLogger(Machine.class);
	
	private RegisterBank registerBank;
	
	private InstructionUnit iu;
	
	private ArrayList<DUnit> dunits;
	
	private Program program;
	
	private ScoreBoard scoreBoard;
	
	// no operation dunit, signifies a dunit that skipped a cycle
	private DUnit noop;
	
	public Machine(InstructionUnit iu, RegisterBank registerBank) {
		this.registerBank = registerBank;
		this.iu = iu;
		this.dunits = new ArrayList<DUnit>();
		this.noop = new DUnit(new Operation(""), 0);
		new ArrayList<Instruction>();
		this.scoreBoard = new ScoreBoard(this.registerBank);
	}
	
	// adds a dunit of operation type
	public void addDUnit(Operation operation, int cycles) {
		DUnit dunit = new DUnit(operation, cycles);
		this.dunits.add(dunit);
		this.scoreBoard.addDUnit(dunit);
	}
	
	// remove all dunits
	public void clearDUnits() {
		this.dunits.clear();
		this.scoreBoard = new ScoreBoard(this.registerBank);
	}
	
	// return a dunit thats not busy and has the right operation
	public DUnit getFreeDUnit(Operation op) {
		for (DUnit d: this.dunits) {
			if (d.isFree() && d.equals(op)) {
				return d;
			}
		}
		return null;
	}
	
	// load a program into memory
	public void loadProgram(Program program) {
		this.program = program;
	}
	
	private void issue(ArrayList<OperableComponent> operations) {

		Instruction inst;
		DUnit dunit;
		

		// once IU is done see if the instruction can be issued
		if (this.iu.isDone()) {

			inst = this.iu.getInstruction();
			
			assert inst != null;
			
			this.noop.start(inst); // set instruction to fix order bug
			operations.add(this.noop);
			
			// detect O-O hazard and do not finish issue
			if (this.scoreBoard.findWAWHazard(inst)) {
				logger.debug("WAW Hazard on: {}", inst.getLine());
				return;
			}
			
			dunit = this.getFreeDUnit(inst.getOperation());
			
			// break out if no DUnit is free
			if (dunit == null) {
				return;
			}
			
			// start the DUnit and reset the IU if there is a free DUnit for this instruction 
			dunit.start(inst);
			
			this.iu.reset();
			
			logger.debug("Activating: {}", dunit);
			this.scoreBoard.activateDUnit(dunit);
		}
		
		// load next instruction is IU is free and there are more instructions
		if (this.iu.isFree() && this.program.hasNext()) {
			inst = this.program.next();
			this.scoreBoard.issueInstruction(inst);
			this.iu.start(inst);
		}
		
		if (this.iu.isBusy()) {
			operations.add(this.iu);
			this.iu.cycle();
		}
	}

	private void readOperands(ArrayList<OperableComponent> operations) {
		
		for (DUnit du: this.dunits) {
			
			// du is free so we can skip it
			if (!du.isReady()) {
				continue;
			}
			
			// check to see if dunit/inst is in conflict
			if (this.scoreBoard.findWARHazard(du)) {
				logger.debug("WAR Hazard found: {}", du);
				this.noop.start(du.getInstruction()); // set instruction to fix order bug
				operations.add(this.noop);
				continue;
			}
			
			du.cycle();
			
		}
	}
	
	private void execute(ArrayList<OperableComponent> operations) {
		for (DUnit du: this.dunits) {
			logger.debug("DUnit: {}, State: {}", du, du.state);
			
			// skip non busy dunits
			if (!du.isBusy()) {
				continue;
			}
			
			this.scoreBoard.executeDUnit(du);
			
			// cycle (execute)
			operations.add(du);
			du.cycle();
		}
	}
	
	private void write(ArrayList<OperableComponent> operations) {
		for (DUnit du: this.dunits) {
			
			// only care about done dunits
			if (!du.isDone()) {
				continue;
			}
			
			// don't write if there is a RAW hazard
			if (this.scoreBoard.findRAWHazard(du)) {
				logger.debug("RAW Hazrd Found: {}", du);
				continue;
			}
			
			this.scoreBoard.writeDUnit(du);
			
		}
	}
		
	// this does all the hard work. Does a cycle of the machine
	public ArrayList<OperableComponent> cycle() {
		ArrayList<OperableComponent> operations = new ArrayList<OperableComponent>();
		
		
		this.issue(operations);
		this.readOperands(operations);
		this.execute(operations);
		this.write(operations);
		
		this.scoreBoard.cycle();
		
		return operations;
	}
	
	// resets the machine back to defaults
	public void reset() {
		for (DUnit du: this.dunits) {
			du.reset();
		}
		this.registerBank.reset();
		this.iu.reset();
		this.program.reset();
	}
	
	// returns the program
	public Program getProgram() {
		return this.program;
	}
	
	public Collection<Register> getRegisters() {
		return this.registerBank.getRegisters();
	}
	
}
