/**
 * CS3725 - Assignment #3
 * Tim Oram (200529220)
 * 
 * License:
 * The MIT License (MIT) Copyright (c) 2013 Tim Oram
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this
 * software and associated documentation files (the "Software"), to deal in the Software
 * without restriction, including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons
 * to whom the Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all copies or
 * substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */
package ca.mitmaro.cs3725.a3.gui;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Map;

import javax.swing.SwingUtilities;
import javax.swing.text.Document;

import ca.mitmaro.cs3725.a3.Application;
import ca.mitmaro.cs3725.a3.Topics;
import ca.mitmaro.cs3725.a3.machine.CycleOperations;
import ca.mitmaro.cs3725.a3.machine.component.Register;
import ca.mitmaro.cs3725.a3.program.Operation;
import ca.mitmaro.prattle.Broker;
import ca.mitmaro.prattle.Listener;

public class GuiState implements Listener {

	private final Broker broker;
	private final Window window;
	private final Document editorDocument;
	
	private boolean editorDirty;
	private boolean machineStarted;
	private boolean machineRunning;
	
	public GuiState(Broker broker, Window window, Application application) {
		this.broker = broker;

		this.broker.registerListener(Topics.Application.reset, this);
		this.broker.registerListener(Topics.Application.programLoad, this);
		this.broker.registerListener(Topics.Application.cycleComplete, this);
		this.broker.registerListener(Topics.Application.programStop, this);
		this.broker.registerListener(Topics.Application.programStart, this);
		this.broker.registerListener(Topics.Application.programRun, this);
		
		this.window = window;
		
		this.editorDocument = this.window.editor.editorDocument;
		
		this.editorDirty = false;
		this.machineStarted = false;
		this.machineRunning = false;
		this.updateGui();
	}

	public void updateGui() {

		boolean optionsEnabled = !this.machineStarted;
		boolean editorEnabled = !this.machineStarted;
		boolean graphEnabled = this.machineStarted;
		
		this.window.options.adders.setEnabled(optionsEnabled);
		this.window.options.adderCycles.setEnabled(optionsEnabled);
		this.window.options.multipliers.setEnabled(optionsEnabled);
		this.window.options.multiplierCycles.setEnabled(optionsEnabled);
		this.window.options.iuCycles.setEnabled(optionsEnabled);
		this.window.options.registers.setEnabled(optionsEnabled);
		this.window.options.cycleTime.setEnabled(optionsEnabled);
		
		this.window.graph.bAuto.setEnabled(graphEnabled && !this.machineRunning);
		this.window.graph.bNext.setEnabled(graphEnabled && !this.machineRunning);
		this.window.graph.bPause.setEnabled(graphEnabled && this.machineRunning);
		this.window.graph.bStop.setEnabled(graphEnabled);

		this.window.editor.bRun.setEnabled(editorEnabled && this.editorDocument.getLength() != 0);
		this.window.editor.bClear.setEnabled(editorEnabled);
		this.window.editor.bOpen.setEnabled(editorEnabled);
		this.window.editor.bSave.setEnabled(editorEnabled && this.editorDirty); // only enable is dirty
		this.window.editor.editor.setEditable(editorEnabled);
		
	}
	
	public void setEditorDirty(boolean value) {
		this.editorDirty = value;
		this.updateGui();
	}
	
	private void applicationReset() {
		this.programStop();
		this.machineStarted = false;
	}

	@SuppressWarnings("unchecked")
	private void programStart(Map<String, Object> context) {
		this.machineStarted = true;
		this.window.graph.graphModel.reset();
		Collection<Register> c = (Collection<Register>)context.get("registers");
		this.window.registers.registerModel.setRegisters(new ArrayList<Register>(c));
	}
	
	private void programRun() {
		this.machineRunning = true;
	}
	
	private void programStop() {
		this.machineRunning = false;
	}

	@SuppressWarnings("unchecked")
	private void applicationNext(Map<String, Object> context) {
		
		final ArrayList<Operation> cycle = new ArrayList<Operation>();
		
		for(CycleOperations.Entry entry: (CycleOperations)context.get("cycle")) {
			while (entry.instruction.getPC() >= cycle.size()) {
				cycle.add(new Operation(""));
			}
			cycle.set(entry.instruction.getPC(), entry.operation);
		}
		
		this.window.graph.graphModel.addCycle(cycle);
	
		for (Register re: (Collection<Register>)context.get("registers")) {
			this.window.registers.registerModel.setRegister(re.getNumber(), re);
		}
		
	}
	
	private void programLoad() {
		this.machineStarted = true;
		this.window.graph.graphModel.reset();
		this.updateGui();
	}
	

	@Override
	public void notify(final String topic, final Map<String, Object> context) {
		
		final GuiState self = this;
		
		// this all has to run on the EDT (or at least parts of it but this makes it easier)
		SwingUtilities.invokeLater(new Runnable() {
			@Override
			public void run() {
				if (topic.equals(Topics.Application.reset)) {
					self.applicationReset();
				} else if (topic.equals(Topics.Application.cycleComplete)) {
					self.applicationNext(context);
				} else if (topic.equals(Topics.Application.programLoad)) {
					self.programLoad();
				} else if (topic.equals(Topics.Application.programStart)) {
					self.programStart(context);
				} else if (topic.equals(Topics.Application.programStop)) {
					self.programStop();
				} else if (topic.equals(Topics.Application.programRun)) {
					self.programRun();
				}
				
				self.updateGui();
			}
		});
	}
	
}
