/**
 * CS3725 - Assignment #3
 * Tim Oram (200529220)
 * 
 * License:
 * The MIT License (MIT) Copyright (c) 2013 Tim Oram
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this
 * software and associated documentation files (the "Software"), to deal in the Software
 * without restriction, including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons
 * to whom the Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all copies or
 * substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */
package ca.mitmaro.cs3725.a3.machine;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.ListIterator;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ca.mitmaro.cs3725.a3.machine.component.DUnit;
import ca.mitmaro.cs3725.a3.machine.component.RegisterBank;
import ca.mitmaro.cs3725.a3.machine.scoreboard.DUnitStatus;
import ca.mitmaro.cs3725.a3.machine.scoreboard.InstructionStatus;
import ca.mitmaro.cs3725.a3.program.Instruction;

public class ScoreBoard implements Cyclable {
	
	private static Logger logger = LoggerFactory.getLogger(ScoreBoard.class);
	
	private ArrayList<InstructionStatus> instructions;
	
	private RegisterBank registers;
	
	private ArrayList<DUnitStatus> dunits;
	
	public ScoreBoard(RegisterBank registers) {
		
		this.dunits = new ArrayList<DUnitStatus>();
		
		this.registers = registers;
		this.instructions = new ArrayList<InstructionStatus>();
	}
	
	public void addDUnit(DUnit dunit) {
		this.dunits.add(new DUnitStatus(dunit));
	}
	
	public void issueInstruction(Instruction inst) {
		this.instructions.add(new InstructionStatus(inst));
	}
	
	public void activateDUnit(DUnit dunit) {
		
		Instruction inst;
		DUnit du;
		DUnit rsDUnit = null;
		DUnit rtDUnit = null;
		
		// create instruction status
		this.getInstructionStatus(dunit.getInstruction()).read();
		DUnitStatus dunitStatus = this.getDUnitStatus(dunit);
		
		int rs = dunit.getInstruction().getRS();
		int rt = dunit.getInstruction().getRT();
		
		ListIterator<DUnitStatus> it = this.dunits.listIterator(this.dunits.size());
		while(it.hasPrevious() && (rsDUnit == null || rtDUnit == null)) {
			
			du = it.previous().dunit;
			
			// skip the active dunit
			if (du == dunit) { continue; }
			
			
			// if the dunit is inactive ignore it
			if (!du.isBusy() && !du.isReady()) {
				continue;
			}
			
			inst = du.getInstruction();
			
			logger.debug("INST RD: {}, DU RS: {}, DU RT", inst.getRD(), rs, rt);
			
			if (rsDUnit == null && inst.getRD() == rs) {
				logger.debug("Setting RS DUnit: {}", du);
				rsDUnit = du;
			}
			
			if (rtDUnit == null && inst.getRD() == rt) {
				logger.debug("Setting RD DUnit: {}", du);
				rtDUnit = du;
			}
		}
		
		dunitStatus.setSourceDUnits(rsDUnit, rtDUnit);
		
		this.registers.getRegister(dunit.getInstruction().getRD()).setDUnit(dunit);
		
	}
	
	public void executeDUnit(DUnit dunit) {
		
		this.getInstructionStatus(dunit.getInstruction()).execute();
		
	}
	
	public void writeDUnit(DUnit dunit) {
		
		// free registers being held by this dunit
		for (DUnitStatus dus: this.dunits) {
			
			if (dus.getRSDUnit() == dunit) {
				dus.setSourceDUnits(null, dus.getRTDUnit());
			}
			
			if (dus.getRTDUnit() == dunit) {
				dus.setSourceDUnits(dus.getRSDUnit(), null);
			}
		}
		
		this.getInstructionStatus(dunit.getInstruction()).done();
		
		// dunit is now free to execute a new instruction
		dunit.reset();
		
	}
	
	@Override
	public void cycle() {

		// clean up any completed instructions
		Iterator<InstructionStatus> it = this.instructions.iterator();
		while(it.hasNext()) {
			if(it.next().isDone()) {
				it.remove();
			}
		}
		
	}


	public boolean findWAWHazard(Instruction inst) {

		for (InstructionStatus is: this.instructions) {
			
			// skip self
			if (is.instruction == inst) {
				continue;
			}
			
			if (is.instruction.getRD() == inst.getRD()) {
				return true;
			}
		}
		
		return false;
	}


	public boolean findWARHazard(DUnit dunit) {
		DUnitStatus dus = this.getDUnitStatus(dunit);
		
		return !dus.rsReady() || !dus.rtReady();
	}


	public boolean findRAWHazard(DUnit dunit) {

		Instruction inst = dunit.getInstruction();
		DUnitStatus dus;
		
		for (InstructionStatus is: this.instructions) {
			
			// exclude instructions that are executing, writing or done
			if (is.isExecuting() || is.isWriting() || is.isDone()) {
				continue;
			}
			
			// don't compare against it's self (this shouldn't happen as instruction should be in write stage)
			if (is.instruction == inst) {
				continue;
			}
			
			is = this.getInstructionStatus(inst);
			dus = this.getDUnitStatus(inst);
			
			
			if (is.instruction.getRS() == inst.getRD() && dus.getRSDUnit() != dunit) {
				return true;
			}
			
			if (is.instruction.getRT() == inst.getRD() && dus.getRTDUnit() != dunit) {
				return true;
			}
		}
		
		return false;
	}

	private DUnitStatus getDUnitStatus(DUnit dunit) {
		for (DUnitStatus dus: this.dunits) {
			if (dus.dunit == dunit) {
				return dus;
			}
		}
		
		throw new RuntimeException("Could not find DUnitStatus for DUnit");
	}


	private DUnitStatus getDUnitStatus(Instruction inst) {
		for (DUnitStatus dus: this.dunits) {
			if (dus.dunit.getInstruction() == inst) {
				return dus;
			}
		}
		
		throw new RuntimeException("Could not find DUnitStatus for DUnit");
	}

	
	private InstructionStatus getInstructionStatus(Instruction inst) {
		for (InstructionStatus is: this.instructions) {
			if (is.instruction == inst) {
				return is;
			}
		}
		
		throw new RuntimeException("Could not find InstructionStatus for Instruction");
		
	}
	
}
