/**
 * The MIT License (MIT) Copyright (c) 2013 Tim Oram
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this
 * software and associated documentation files (the "Software"), to deal in the Software
 * without restriction, including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons
 * to whom the Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all copies or
 * substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

package ca.mitmaro.prattle;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * A concurrent message broker 
 * 
 * @author Tim Oram
 */
public class Broker {
	
	private static Logger logger = LoggerFactory.getLogger(Broker.class);
	
	/**
	 * Internal handler for executor services, wraps a message listener.
	 * 
	 * @author Tim Oram
	 */
	private static class Handler implements Runnable {
		
		/** The topic */
		private String topic;
		
		/** The listener */
		private Listener listener;
		
		/** The context */
		private Map<String, Object> context;
		
		/**
		 * Construct a handler with a topic and listener
		 * @param topic
		 * @param listener 
		 */
		public Handler(String topic, Map<String, Object> context, Listener listener) {
			
			this.topic = topic;
			this.listener = listener;
			this.context = context;
			
		}

		/**
		 * Runs the listener updated method
		 */
		public void run() {
			
			this.listener.notify(this.topic, this.context);
			
		}
		
	}
	
	/** A map of topics to a list of listeners */
	private Map<String, List<Listener>> listeners;
	
	/** The executor service used for messaging listeners concurrently */
	private ExecutorService executors;
	
	/**
	 * Construct a broker using a default map and executor service
	 */
	public Broker() {
		this(new HashMap<String, List<Listener>>(), Executors.newCachedThreadPool());
	}
	
	/**
	 * Construct a broker using the provided map and a default executor service
	 * @param topic_listners
	 */
	public Broker(Map<String, List<Listener>> topic_listners) {
		this(topic_listners, Executors.newCachedThreadPool());
	}
	
	/**
	 * Construct a broker using the provided executor service and a default map
	 * @param executor
	 */
	public Broker(ExecutorService executor) {
		this(new HashMap<String, List<Listener>>(), executor);
	}
	
	/**
	 * Construct a broker using the provided executor service and map
	 * @param listeners
	 * @param executors
	 */
	public Broker(Map<String, List<Listener>> listeners, ExecutorService executors) {
		this.listeners = listeners;
		this.executors = executors;
	}
	
	/**
	 * Register a topic providing a list for the listeners
	 * 
	 * This method will remove any existing listeners registered to a topic
	 * @param topic
	 * @param listeners
	 */
	public void registerTopic(String topic, List<Listener> listeners) {
		
		// register
		this.listeners.put(topic, listeners);
		
	}
	
	/**
	 * Register a topic with no listeners. Multiple calls to this method will have no effect.
	 * @param topic
	 */
	public void registerTopic(String topic) {
		
		// already registered
		if (this.listeners.containsKey(topic)) {
			return;
		}
		
		// register
		this.registerTopic(topic, new LinkedList<Listener>());
		
	}
	
	/**
	 * Register an array of topics.
	 * @param topics
	 */
	public void registerTopic(String[] topics) {
		
		for (String topic: topics) {
			this.registerTopic(topic);
		}
		
	}
	
	/**
	 * Register a listener to the provided topic. 
	 * @param topic
	 * @param listener
	 */
	public void registerListener(String topic, Listener listener) {
		
		this.getListeners(topic).add(listener);
		
	}
	
	/**
	 * Registers an array of listeners to the provided topic.
	 * @param topic
	 * @param listeners
	 */
	public void registerListener(String topic, Listener[] listeners) {
		
		for (Listener listener: listeners) {
			this.registerListener(topic, listener);
		}
		
	}

	/**
	 * Remove a listener from the provided topic.
	 * @param topic
	 * @param listener
	 */
	public void removeListener(String topic, Listener listener) {
		
		this.getListeners(topic).remove(listener);
		
	}
	
	/**
	 * Remove a topic, will also remove all registered listeners.
	 * @param topic
	 */
	public void removeTopic(String topic) {
		
		this.listeners.remove(topic);
		
	}
	
	/**
	 * Post the topic to the broker, the broker will notify listeners
	 * 
	 * @param topic
	 */
	public void post(String topic) {
		this.post(topic, null);
	}
	
	/**
	 * Post the topic to the broker with a context, the broker will notify listeners
	 * @param topic
	 * @param context
	 */
	public void post(String topic, Map<String, Object> context) {
		
		logger.trace("Posting: {}", topic);
		
		for (Listener listener: this.getListeners(topic)) {
			this.executors.execute(new Handler(topic, context, listener));
		}
		
	}
	
	/**
	 * Get listeners for the provided topic. The topic must have been registered previously.
	 * @param topic
	 * @return The listeners for the provided topic
	 */
	private List<Listener> getListeners(String topic) {
		
		if (!this.listeners.containsKey(topic)) {
			throw new RuntimeException(String.format("%s is not a registered topic", topic));
		}
		
		return this.listeners.get(topic);
		
	}
	
}
